(* 
app load ["Int", "TUjjak", "Timer"];
load "Ujjak";
*)

structure KUjjak :> KUjjak =
    (* dp04a-nhf v1.0 -- 2004. november 1. Hank Pter & Dvid *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KUjjak of string

(* helyes_feladvany tabla = tabla egy helyes szerkezet feladvny.
   helyes_feladvany : TUjjak.ujjtabla -> bool
 *)
fun helyes_feladvany tabla =
    let
	fun helyes_mezo v = 0 <= v andalso v <= 9

	val n = length tabla
    in
	(* ngyzet alak *)
	List.all (fn ls => n = length ls) tabla andalso
	(* nem res *)
	n <> 0 andalso
	(* minden mezje helyes *)
	List.all (List.all helyes_mezo) tabla
    end

local
    fun elementOf [] _      = false
      | elementOf (x::xs) y = y = x orelse elementOf xs y

    fun contains_only (xs,is) = List.all (elementOf is) xs

    open TUjjak
in
    (* helyes_megoldas N u = u egy N mret feladvny helyes
                             szerkezet megoldsa
       helyes_megoldas int -> TUjjak.peremezes -> bool
     *)
    fun helyes_megoldas N (na,ea,sa,wa) =
	(* helyes mret *)
	N = length na andalso N = length ea    andalso
	N = length sa andalso N = length wa    andalso
        (* nem mutat kifel *)
	contains_only (na, [sw, s, se])        andalso
	contains_only (ea, [nw, w, sw])        andalso
	contains_only (sa, [nw, n, ne])        andalso
	contains_only (wa, [ne, e, se])        andalso
	hd na <> sw andalso List.last na <> se andalso
	hd ea <> nw andalso List.last ea <> sw andalso
	hd sa <> nw andalso List.last sa <> ne andalso
	hd wa <> ne andalso List.last wa <> se
end


(* ujjak_be f = az f szveges llomnybl beolvasott feladvny.
   ujjak_be : string -> TUjjak.ujjtabla
 *)
fun ujjak_be f =
    let
	fun str2field c =
	    case Int.fromString c of
		 NONE   => raise KUjjak("Nemvrt karaktersorozat: '" ^ c ^ "'.")
	       | SOME v => if 0 <= v andalso v <= 9 then v
			   else raise KUjjak("Hibs mezrtk: '" ^ c ^ "'.")

	val istr  = TextIO.openIn f
	val all   = TextIO.inputAll istr before TextIO.closeIn istr
	val lines = String.tokens (fn c => c = #"\n") all
	val tabla = map (map str2field o String.tokens Char.isSpace) lines
    in
	if helyes_feladvany tabla then tabla
	else raise KUjjak("Hibs feladvny a(z) " ^ f ^ " llomnyban.")
    end

(* arrow2str u = az u ujj fzr alakja
   arrow2str : TUjjak.ujj -> string
 *)
local open TUjjak
in
    fun arrow2str s  = "|"
      | arrow2str sw = "/"
      | arrow2str w  = "-"
      | arrow2str nw = "\\"
      | arrow2str n  = "|"
      | arrow2str ne = "/"
      | arrow2str e  = "-"
      | arrow2str se = "\\"
end

(* field2str f = az f mez fzr alakja
   field2str : TUjjak.mezo -> string
 *)
fun field2str v = Int.toString v

(* pretty_print tabla peremezes = a tbla peremezse "szp" formban
   pretty_print : TUjjak.ujjtabla * TUjjak.peremezes -> string
 *)
fun pretty_print (t, (n,e,s,w)) =
    let
	val _ = helyes_megoldas (length t) (n,e,s,w) orelse
	        raise KUjjak "Hibs szerkezet megolds!"
	(* arrow_line : TUjjak.ujjak -> string *)
	val arrow_line = foldl (fn (a,s) => s ^ "   " ^ arrow2str a) " "
	val seprow     = "  " ^ foldl (fn (_,s) => s ^ "---+") "+" (hd t)
	(* table_line : TUjjak.sor -> string *)
	val table_line = foldl (fn (f,s) => s ^ field2str f ^ " | ") " | "

	fun lines2str ([],[],[]) = seprow
	  | lines2str (w::ws, l::ls, e::es) =
	      seprow ^ "\n" ^
	      arrow2str w ^ table_line l ^ arrow2str e ^ "\n" ^
	      lines2str (ws,ls,es)
	  | lines2str _ = raise Match
    in
	arrow_line n ^ "\n" ^ lines2str(w,t,e) ^ "\n" ^ arrow_line s ^ "\n"
    end

(* ujjak_ki(m, t, us) = a t feladvny us megoldslistjt olvashat
                        formban kirja az m szveges llomnyba.
   ujjak_ki : string * TUjjak.ujjtabla * TUjjak.peremezes list -> unit
*)
fun ujjak_ki(m, t, us) =
      let fun pp (u,s) = s ^ (pretty_print (t,u) handle KUjjak s => s) ^ "\n"
	  val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, foldl pp "" us)
      in
	 TextIO.closeOut ostr
      end
  
(* megold(f, m) = beolvas egy feladvnyt az f szveges llomnybl s
                  kirja az sszes megoldst az m szveges llomnyba.
                  Ehhez termszetesen szksge van a Ujjak.ujjak
                  fggvnyre.  Eredmnye az f nevet, a megoldsok szmt
                  s a futsi idt msodpercben tartalmaz fzr.
   megold : string * string -> string
*)
fun megold (f, m) =
    let val t = ujjak_be f
	val start = Timer.startCPUTimer()
	val us = Ujjak.ujjak t
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = ujjak_ki(m, t, us)
    in
	"Ler: " ^ f ^ ", " ^
	Int.toString(length us) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end

end
